import Foundation

// On the limitations of CharacterSet
// **********************************

// Task: Find all the characters that match a list of invalid characters and remove them form the input. Return those removed party joined as a single string or nil if the input was valid.

let input = "1234<>.567;"
let forbiddenString = "<>.;*"
let forbiddenCharacterSet = CharacterSet(charactersIn: forbiddenString)

// Nomenclature: n = |input|, k = |forbiddenCharacters|

func sanitize1(input: String) -> (removed: String?, sanitized: String) {
    var removed: String?
    let sanitized = input.filter { (character) -> Bool in //O(n)
        if forbiddenCharacterSet.contains(character) { //O(1) (Set containment is constant)
            
            // Store the removed character
            let forbiddenCharacterAsString = String(character)
            removed = (removed ?? "") // Inizializing it only when needed
            removed?.append(forbiddenCharacterAsString)
            return false
        }
        return true
    }
    return (removed, sanitized)
    // O(n), BUT: "Cannot convert value of type 'String.Element' (aka 'Character') to expected argument type 'Unicode.Scalar'"
}

func sanitize1_1(input: String) -> (removed: String?, sanitized: String) {
    var removed: String?
    let sanitized = input.filter { (character) -> Bool in //O(n)
        
        for scalar in character.unicodeScalars { //O(|unicode scalars in character|)
            if forbiddenCharacterSet.contains(scalar) { //O(1) (Set containment is constant)
                // Store the removed character
                let forbiddenCharacterAsString = String(character)
                removed = (removed ?? "") // Inizializing it only when needed
                removed?.append(forbiddenCharacterAsString)
                return false
            }
        }
        return true
    }
    return (removed, sanitized)
    // O(n) * O(|unicode scalars in character|)
}

func sanitize2(input: String) -> (removed: String?, sanitized: String) {
    var removed: String?
    var sanitized = input
    
    let inputSet = CharacterSet(charactersIn: input) // O(n)
    let intersection = inputSet.intersection(forbiddenCharacterSet) // O(min(n,k))
    if !intersection.isEmpty {
        // Input contains invalid characters, only perform the filter now
        sanitized = input.filter { (character) -> Bool in //O(n)
            if forbiddenCharacterSet.contains(character) { //O(1) (Set containment is constant)
                
                // Store the removed character
                let forbiddenCharacterAsString = String(character)
                removed = (removed ?? "") // Inizializing it only when needed
                removed?.append(forbiddenCharacterAsString)
                return false
            }
            return true
        }
    }
    return (removed, sanitized)
    // Filter only performed when needed, but checking for containment is not possible.
    // ALSO: Getting characters OUT of a CharacterSet is not possible
    // O(n) + O(min(n,k)), when it contains invalid characters, O(min(n,k)) if not
}

func sanitize3(input: String) -> (removed: String?, sanitized: String) {
    var removed: String?
    // Using string containment now.
    let sanitized = input.filter { (character) -> Bool in //O(n)
        
        if forbiddenString.contains(character) { // O(k), not a set operation
            // Store the removed character
            let forbiddenCharacterAsString = String(character)
            removed = (removed ?? "") // Inizializing it only when needed
            removed?.append(forbiddenCharacterAsString)
            return false
        }
        return true
    }
   
    return (removed, sanitized)
    // O(k*n) in total
}

func sanitize4(input: String) -> (removed: String?, sanitized: String) {
    var removed: String?
    var sanitized = input
    
    // Documentation: "This method does not perform any Unicode normalization."
    while let range = sanitized.rangeOfCharacter(from: forbiddenCharacterSet) { // O(n)*k
        let wrongPart = String(sanitized[range])
        removed = (removed ?? "") // Inizializing it only when needed
        removed?.append(wrongPart)
        sanitized.removeSubrange(range)
    }
    
    return (removed, sanitized)
    // O(k*n) in total but saves a little because .rangeOfCharacter() fill find elements more quickly.
}



let result = sanitize4(input: input)
if let forbidden = result.removed {
    print("🙅🏻‍♂️ The characters \(forbidden) were removed.")
} else {
    print("Everything was valid 👍🏼")
}
print("Sanitized result: \(result.sanitized)")
